var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
import { invariant } from 'outvariant';
import { parse } from 'set-cookie-parser';
import { Headers, headersToObject } from 'headers-utils';
import { rest, context, } from 'msw';
import { decodeBase64String } from './utils/decodeBase64String';
const defaultResponseProducer = (res, transformers) => {
    return res(...transformers);
};
/**
 * Create a request handler from the given Network Archive entry.
 */
function toRequestHandler(entry, produceResponse = defaultResponseProducer) {
    const { request } = entry;
    const method = request.method.toLowerCase();
    const transformers = toResponseTransformers(entry);
    return rest[method](request.url, (req, res) => {
        return produceResponse(res, transformers);
    });
}
/**
 * Map a traffic entry to the array of response transformers.
 */
export function toResponseTransformers(entry) {
    const { response, time } = entry;
    const transformers = [];
    const responseHeaders = new Headers();
    const responseCookies = [];
    // Response status and status text.
    transformers.push(context.status(response.status, response.statusText));
    // Response headers.
    for (const header of response.headers) {
        const headerName = header.name.toLowerCase();
        // Skip response cookie headers because a mocked response cookies
        // are not implemented through headers (security consideration).
        // Store the list of cookie headers to apply them via `ctx.cookie` later.
        if (['set-cookie', 'set-cookie2'].includes(headerName)) {
            responseCookies.push(...parse(header.value));
            continue;
        }
        // Skip the "Content-Encoding" header to prevent "incorrect header check" errors.
        // MSW must not attempt to compress the response body, even if it was originally
        // compressed. All response bodies are sent uncompressed.
        if (headerName === 'content-encoding') {
            continue;
        }
        responseHeaders.set(header.name, header.value);
    }
    transformers.push(context.set(headersToObject(responseHeaders)));
    // Response cookies.
    for (const cookie of responseCookies) {
        const { name, value } = cookie, options = __rest(cookie, ["name", "value"]);
        transformers.push(context.cookie(name, value, Object.assign(Object.assign({}, options), { sameSite: options.sameSite === '' })));
    }
    // Response delay.
    if (time) {
        transformers.push(context.delay(time));
    }
    // Response body.
    const responseBody = toResponseBody(response);
    if (responseBody) {
        transformers.push(context.body(responseBody));
    }
    return transformers;
}
/**
 * Extract a response body from the given HAR response entry.
 * Decodes any base64-encoded text response bodies.
 */
export function toResponseBody(response) {
    const { text, encoding, mimeType } = response.content;
    if (!text) {
        return;
    }
    if (encoding === 'base64' && mimeType.includes('text')) {
        const responseBody = decodeBase64String(text);
        return responseBody;
    }
    return text;
}
/**
 * Generate request handlers from the given HAR file.
 */
export function fromTraffic(har, mapEntry) {
    invariant(har, 'Failed to generate request handlers from traffic: expected an HAR object but got %s.', typeof har);
    invariant(har.log.entries.length > 0, 'Failed to generate request handlers from traffic: given HAR object has no entries.');
    const requestPaths = new Set();
    const handlers = har.log.entries.reduceRight((handlers, entry) => {
        const resolvedEntry = mapEntry ? mapEntry(entry) : entry;
        if (!resolvedEntry) {
            return handlers;
        }
        const requestPath = createRequestPath(resolvedEntry.request);
        const isUniqueHandler = !requestPaths.has(requestPath);
        const handler = toRequestHandler(resolvedEntry, (res, transformers) => {
            // Reducing the entries from right to left implies that the first
            // entry we meet is, in fact, the last entry recorded.
            // Always create a regular handler for the last entry.
            // If there are any consecutive entries for the same URL,
            // create a one-time handler instead to preserve response order.
            const responseFn = isUniqueHandler ? res : res.once;
            return responseFn(...transformers);
        });
        // Prepend the handler to the list of handler because we're reducing
        // from right to left, but the order of handlers must correspond
        // to the chronological order of requests.
        handlers.unshift(handler);
        requestPaths.add(requestPath);
        return handlers;
    }, []);
    return handlers;
}
function createRequestPath(request) {
    return `${request.method}+${request.url}`;
}
